@@ -192,42 +192,41 @@ def generate_pheno_plink_fast_do(
     outdir: str,
 ) -> pd.DataFrame:
     """
-    DO behavior (mirror Hao's 'J:DO' branch, but streaming the PED):
+    DO behavior (matches Hao):
+      - animal-level rows keyed by animal_id (no genotype duplication)
+      - FID=V1 and IID=V2 from the DO PED, SEX set from pheno, PHE=zscore
+      - PHENO: FID=V1, IID=V2, zscore, value
+      - Order follows DO PED order
 
-      - Find animals where pheno.animal_id overlaps PED V1 (after ID normalization).
-      - For each overlapping PED row, attach sex/measnum/zscore/value from pheno.
-      - For each (measnum, sex), write:
-          * PED: V1..last_col (meta + genotypes), with V5 = sex, V6 = zscore
-          * MAP: full map with '.' -> 'chr_bp'
-          * PHENO: FID = V1, IID = V2, zscore, value
-
-      Implementation avoids loading the PED into memory; it scans ped_file once.
+    IMPORTANT:
+      - We do *not* pandas-read the PED; we stream it line-by-line.
+      - We fail loudly if there is ZERO overlap between pheno animal_id
+        (normalized) and PED V1 (normalized).
     """
     os.makedirs(outdir, exist_ok=True)
     if pheno is None or pheno.empty:
         return pd.DataFrame()
 
-    # Required columns
     need = ("strain", "sex", "measnum", "value", "animal_id")
     missing = [c for c in need if c not in pheno.columns]
     if missing:
-        raise ValueError(f"[DO] pheno missing required columns: {missing} (need {list(need)})")
+        raise ValueError(f"[generate_pheno_plink_fast_do] pheno missing required columns: {missing} (need {list(need)})")
 
     ph = pheno.copy()
-    # Match Hao: strip spaces from strain
+
+    # Normalize strain a bit (Hao just checks for "J:DO" membership)
     ph["strain"] = ph["strain"].astype(str).str.replace(" ", "", regex=False)
-    # Only keep f/m
+
+    # Keep only usable rows (sex f/m, non-empty animal_id)
     ph = ph[ph["sex"].isin(["f", "m"])].copy()
+    ph["animal_id_norm"] = ph["animal_id"].map(_norm_id)
+    ph = ph[ph["animal_id_norm"].notna() & (ph["animal_id_norm"] != "")].copy()
+
     if ph.empty:
+        logging.info("[generate_pheno_plink_fast_do] No usable rows after filtering sex/animal_id; nothing to write.")
         return ph
 
-    # Hao's DO branch is triggered when strain == "J:DO"
-    if "J:DO" not in set(ph["strain"]):
-        logging.warning("[DO] No 'J:DO' strain in pheno; falling back to NON-DO behavior.")
-        # IMPORTANT: NON-DO path is already validated and should not be touched
-        return generate_pheno_plink_fast_non_do(ped_file, map_file, pheno, outdir)
-
-    # MAP sanitize (same as NON-DO / Hao)
+    # MAP sanitize (same as NON-DO)
     map_df = pd.read_csv(map_file, header=None, sep=r"\s+", engine="python")
     map_df[1] = np.where(
         map_df[1].astype(str) == ".",
@@ -235,149 +234,154 @@ def generate_pheno_plink_fast_do(
         map_df[1].astype(str),
     )
 
-    # Ensure zscore column exists
+    # zscore presence
     if "zscore" not in ph.columns:
-        logging.info("[DO] 'zscore' missing; filling NaN (becomes -9).")
+        logging.info("[generate_pheno_plink_fast_do] 'zscore' missing; filling NaN (becomes -9).")
         ph["zscore"] = np.nan
 
-    # ------------------------------------------------------------------
-    # Build per-animal lookup from pheno, mimicking:
-    #   pheno.overlap.id =
-    #     pheno %>%
-    #       filter(animal_id %in% overlap.id) %>%
-    #       mutate(animal_id = as.numeric(animal_id)) %>%
-    #       slice(match(ped.overlap.id$V1, animal_id))
-    #
-    # We normalize IDs with _norm_id and keep the *first* row per animal_id,
-    # just like match() would.
-    # ------------------------------------------------------------------
-    ph["animal_id_norm"] = ph["animal_id"].map(_norm_id)
+    # Build per-(measnum,sex): animal_id_norm -> (zscore, value, sex)
+    per_group_maps: Dict[tuple, Dict[str, tuple]] = {}
+    global_ids: set[str] = set()
 
-    per_id: Dict[str, tuple] = {}
-    for _, r in ph.iterrows():
-        aid = r["animal_id_norm"]
-        if not aid:
-            continue
-        # First occurrence wins (match-like behavior)
-        if aid in per_id:
-            continue
-        # Coerce measnum, zscore, value to numeric-ish
-        try:
-            meas = int(r["measnum"])
-        except Exception:
-            # if something insane comes through, just skip it
-            continue
-        sex = str(r["sex"])
-        z = r.get("zscore", np.nan)
-        v = r.get("value", np.nan)
-        try:
-            z = float(z)
-        except Exception:
-            z = np.nan
-        try:
-            v = float(v)
-        except Exception:
-            v = np.nan
-        per_id[aid] = (sex, meas, z, v)
+    for (meas, sex), g in ph.groupby(["measnum", "sex"], sort=False):
+        meas = int(meas)
+        sex = str(sex)
 
-    if not per_id:
-        logging.info("[DO] No overlapping animal IDs between pheno and PED; nothing to do.")
-        return ph.iloc[0:0].copy()
-
-    # ------------------------------------------------------------------
-    # Stream PED once, in file order, and for each overlapping V1:
-    #   - update V5/V6 from pheno (sex/zscore)
-    #   - route the row into the correct (measnum, sex) PED + PHENO files
-    #   - write MAP once per (measnum, sex)
-    # This reproduces Hao's ped.overlap.id + pheno.ped behavior but without
-    # loading the PED into memory.
-    # ------------------------------------------------------------------
-    handles: Dict[tuple, tuple] = {}  # (meas, sex) -> (ped_fh, phe_fh)
-    wrote_map_for: set[tuple] = set()  # (meas, sex) that already have a .map
-
-    with open(ped_file, "r", encoding="utf-8", errors="replace") as fped:
-        for raw in fped:
-            if not raw.strip():
-                continue
-            # Split on any whitespace, like PLINK and fread() do
-            parts = raw.rstrip("\n").split()
-            if len(parts) < 7:
-                # malformed PED line, ignore (Hao's fread would drop these)
+        m: Dict[str, tuple] = {}
+        seen: set[str] = set()
+        for _, r in g.iterrows():
+            aid = r["animal_id_norm"]
+            if aid in seen:
+                # Exactly like Hao's slice(match(...)): keep first record per animal
                 continue
+            seen.add(aid)
 
-            V1, V2 = parts[0], parts[1]
-            aid_norm = _norm_id(V1)
+            z = r.get("zscore", np.nan)
+            v = r.get("value", np.nan)
+            try:
+                z = float(z)
+            except Exception:
+                z = np.nan
+            try:
+                v = float(v)
+            except Exception:
+                v = np.nan
 
-            if aid_norm not in per_id:
-                # This PED animal is not in the pheno set
-                continue
+            m[aid] = (z, v, sex)
+            global_ids.add(aid)
 
-            sex_label, measnum, z, v = per_id[aid_norm]
-            key = (measnum, sex_label)
+        if m:
+            per_group_maps[(meas, sex)] = m
 
-            # Lazily open PED/PHENO/MAP for this (measnum, sex)
-            if key not in handles:
-                ped_out = os.path.join(outdir, f"{measnum}.{sex_label}.ped")
-                phe_out = os.path.join(outdir, f"{measnum}.{sex_label}.pheno")
+    if not per_group_maps:
+        logging.info("[generate_pheno_plink_fast_do] No per-(measnum,sex) groups after dedup; nothing to write.")
+        return ph
 
-                ped_fh = open(ped_out, "w", encoding="utf-8")
-                phe_fh = open(phe_out, "w", encoding="utf-8")
-                handles[key] = (ped_fh, phe_fh)
+    # Quick sanity log
+    logging.info(
+        "[generate_pheno_plink_fast_do] groups=%d, total unique animals=%d",
+        len(per_group_maps),
+        len(global_ids),
+    )
 
-                # Write MAP once for this (measnum, sex)
-                if key not in wrote_map_for:
-                    map_out = os.path.join(outdir, f"{measnum}.{sex_label}.map")
-                    map_df.to_csv(map_out, sep="\t", header=False, index=False)
-                    wrote_map_for.add(key)
-                    logging.info(
-                        "[generate_pheno_plink_fast:DO] wrote %s", map_out
+    # Prepare output handles per group
+    group_handles: Dict[tuple, dict] = {}
+    for (meas, sex), aid_map in per_group_maps.items():
+        map_out = os.path.join(outdir, f"{meas}.{sex}.map")
+        ped_out = os.path.join(outdir, f"{meas}.{sex}.ped")
+        phe_out = os.path.join(outdir, f"{meas}.{sex}.pheno")
+
+        # Write MAP once per group
+        map_df.to_csv(map_out, sep="\t", index=False, header=False)
+
+        group_handles[(meas, sex)] = {
+            "aid_map": aid_map,
+            "ped_path": ped_out,
+            "phe_path": phe_out,
+            "ped_file": open(ped_out, "w", encoding="utf-8"),
+            "phe_file": open(phe_out, "w", encoding="utf-8"),
+            "wrote_any": False,
+        }
+
+    # Single streaming pass per group would re-scan PED for each group.
+    # Instead, we do ONE pass over PED and fan out to groups that contain this animal.
+    # This keeps order == PED order, like Hao's ped.overlap.id.
+    try:
+        with open(ped_file, "r", encoding="utf-8", errors="replace") as fped:
+            for raw in fped:
+                if not raw.strip():
+                    continue
+                parts = raw.rstrip("\n").split()
+                if len(parts) < 7:
+                    # malformed / header-like, skip
+                    continue
+
+                V1, V2 = parts[0], parts[1]
+                V1n = _norm_id(V1)
+
+                if V1n not in global_ids:
+                    # This animal never appears in any (meas, sex) group
+                    continue
+
+                # For each group that has this animal, write appropriate lines
+                for key, info in group_handles.items():
+                    aid_map = info["aid_map"]
+                    if V1n not in aid_map:
+                        continue
+
+                    z, v, sx = aid_map[V1n]
+
+                    # Copy meta so we don't mutate 'parts' between groups
+                    meta = parts[:6]
+                    # SEX: 2=f, 1=m
+                    meta[4] = "2" if sx == "f" else "1"
+                    # PHE: zscore or -9
+                    meta[5] = f"{z}" if math.isfinite(z) else "-9"
+
+                    # Reconstruct PED line: meta + genotype pairs as-is
+                    out_line = " ".join(meta + parts[6:])
+                    info["ped_file"].write(out_line + "\n")
+
+                    # PHENO: FID V1, IID V2, z, value
+                    info["phe_file"].write(
+                        f"{V1} {V2} "
+                        f"{(z if math.isfinite(z) else -9)} "
+                        f"{(v if math.isfinite(v) else -9)}\n"
                     )
+                    info["wrote_any"] = True
+    finally:
+        # Close all files
+        for info in group_handles.values():
+            info["ped_file"].close()
+            info["phe_file"].close()
 
-            ped_fh, phe_fh = handles[key]
+    # Log + sanity: did we write anything at all?
+    any_written = any(info["wrote_any"] for info in group_handles.values())
 
-            # Update V5 (SEX) and V6 (PHE) like Hao:
-            #   V5 = 2 if f, 1 if m
-            #   V6 = zscore, with missing -> -9
-            parts_meta = parts[:6]
-            if sex_label == "f":
-                parts_meta[4] = "2"
-            elif sex_label == "m":
-                parts_meta[4] = "1"
-            # else: just leave whatever is there (shouldn't happen)
-
-            if math.isfinite(z):
-                parts_meta[5] = f"{z}"
-                z_out = z
-            else:
-                parts_meta[5] = "-9"
-                z_out = -9.0
-
-            # The rest of the line are genotype tokens already split (V7+),
-            # so we just preserve them.
-            parts_out = parts_meta + parts[6:]
-
-            ped_fh.write(" ".join(parts_out) + "\n")
-
-            # PHENO: FID = V1, IID = V2, zscore, value
-            if not math.isfinite(v):
-                v_out = -9.0
-            else:
-                v_out = v
-            phe_fh.write(f"{V1} {V2} {z_out} {v_out}\n")
-
-    # Close all open handles
-    for ped_fh, phe_fh in handles.values():
-        ped_fh.close()
-        phe_fh.close()
-
-    # Log what we actually produced
-    if not handles:
-        logging.info("[generate_pheno_plink_fast:DO] No DO PED/PHENO files were written.")
-    else:
-        for (measnum, sex_label) in sorted(handles.keys()):
+    for (meas, sex), info in group_handles.items():
+        if info["wrote_any"]:
             logging.info(
-                "[generate_pheno_plink_fast:DO] finished measnum=%s sex=%s", measnum, sex_label
+                "[generate_pheno_plink_fast_do] wrote %s, %s, %s",
+                info["ped_path"],
+                os.path.join(outdir, f"{meas}.{sex}.map"),
+                info["phe_path"],
             )
+        else:
+            # remove empty files to avoid confusing downstream PLINK2
+            try:
+                if os.path.getsize(info["ped_path"]) == 0:
+                    os.remove(info["ped_path"])
+                if os.path.getsize(info["phe_path"]) == 0:
+                    os.remove(info["phe_path"])
+            except OSError:
+                pass
+
+    if not any_written:
+        raise ValueError(
+            "[generate_pheno_plink_fast_do] No DO PED/PHENO files were written because there is no overlap "
+            "between pheno['animal_id'] (normalized) and PED V1 (normalized). "
+            "Check that the DO PED file actually uses the same animal IDs as the measure CSV."
+        )
 
     return ph
 
@@ -404,11 +408,13 @@ def generate_pheno_plink_fast(
     if pt == "DO":
         logging.info("[generate_pheno_plink_fast] using DO panel_type")
         return generate_pheno_plink_fast_do(ped_file, map_file, pheno, outdir)
-    elif pt == "NON_DO":
-        logging.info("[generate_pheno_plink_fast] using NON-DO panel_type")
-        return generate_pheno_plink_fast_non_do(ped_file, map_file, pheno, outdir)
-    else:
-        raise ValueError(f"panel_type must be 'DO' or 'NON_DO', got {panel_type!r}")
+
+    # Everything else is NON-DO
+    logging.info(
+        "[generate_pheno_plink_fast] panel_type=%r => NON-DO path",
+        panel_type,
+    )
+    return generate_pheno_plink_fast_non_do(ped_file, map_file, pheno, outdir)
 
 
 # ----------------------------- Orchestrator ---------------------------- #
